/*
 *   Copyright (c) 2000 S. Noda (HZG02104@nifty.ne.jp)
 *
 *   Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose is hereby granted provided
 * that the above copyright notice and this permission notice appear
 * in all copies of the software and related documentation.
 *
 *                          NO WARRANTY
 *
 *   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY WARRANTIES;
 * WITHOUT EVEN THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
 * FOR A PARTICULAR PURPOSE.
 */

#include <crtdbg.h>
#include <windows.h>
#include <stdio.h>
#include "cmdline.h"
#include "zipdll.h"

static HINSTANCE hDll;

static ZP_VERSION fpZpVersion;
static ZP_INIT fpZpInit;
static ZP_SETOPTIONS fpZpSetOptions;
static ZP_GETOPTIONS fpZpGetOptions;
static ZP_ARCHIVE fpZpArchive;

static ZIPUSERFUNCTIONS init;
static ZPOPT options;
static ZCL archive;

static char *password;
static char *sfxdat;
static int FNV_max;

int WINAPI ZpPrint(LPSTR buf, unsigned long size)
{
	printf ("%s", buf);
	return (unsigned int) size;
}

int WINAPI ZpPassword(LPSTR p, int n, LPCSTR m, LPCSTR name)
{
	if (!password)
	{
		fprintf (stderr, "%s", m);
		scanf ("%s", p);
	}
	else
		strcpy (p, password);
	return 0;
}

int WINAPI ZpService(LPCSTR filename, unsigned long size)
{
	_RPTF2 (_CRT_WARN, "%s (%d)\n", filename, size);
	return FALSE;
}

int WINAPI ZpComment(LPSTR buf)
{
	fprintf (stderr, "enter new zip file comment: ");
	scanf ("%s", buf);
	return TRUE;
}

BOOL ZpLoadLibrary()
{
	hDll = LoadLibrary ("zip32.dll");
	
	if (hDll)
	{
		fpZpVersion = (ZP_VERSION) GetProcAddress (hDll, "ZpVersion");
		fpZpInit = (ZP_INIT) GetProcAddress (hDll, "ZpInit");
		fpZpSetOptions = (ZP_SETOPTIONS) GetProcAddress (hDll, "ZpSetOptions");
		fpZpGetOptions = (ZP_GETOPTIONS) GetProcAddress (hDll, "ZpGetOptions");
		fpZpArchive = (ZP_ARCHIVE) GetProcAddress (hDll, "ZpArchive");
		
		if (fpZpVersion && fpZpInit && fpZpSetOptions && fpZpGetOptions && fpZpArchive)
			return TRUE;
	}
	
	return FALSE;
}

BOOL ZpCheckLibrary()
{
	ZpVer ver;
	(*fpZpVersion) (&ver);
	
	if (ver.zip.major == 2 && ver.zip.minor == 3)
		return TRUE;
	
	return FALSE;
}

int ZpInitLibrary()
{
	init.print = ZpPrint;
	init.comment = ZpComment;
	init.password = ZpPassword;
	init.ServiceApplication = ZpService;
	
	return (*fpZpInit) (&init);
}

void ZpFreeLibrary()
{
	if (hDll)
		FreeLibrary (hDll);
	
	hDll = NULL;
	
	fpZpVersion = NULL;
	fpZpInit = NULL;
	fpZpSetOptions = NULL;
	fpZpGetOptions = NULL;
	fpZpArchive = NULL;
}

void ZpDiscardCmdLine()
{
	sfxdat = NULL;
	options.szTempDir = NULL;
	password = NULL;
	options.Date = NULL;
	options.szRootDir = NULL;
	archive.lpszZipFN = NULL;
	for (int i = 0; i < archive.argc; i++)
		archive.FNV[i] = NULL;
	
	if (archive.FNV)
		free (archive.FNV);
	archive.argc = 0;
	archive.FNV = NULL;
	FNV_max = 0;
}

BOOL ZpParseCmdLine(int ac, char **av)
{
	for (int i = 0; i < ac; i++)
	{
		if (*av[i] == '-')
		{
			char *opt = av[i] + 1;
			if (strncmp (opt, "-sfx", 4) == 0)
			{
				char *p = opt + 4;
				if (*p == '\0')
					sfxdat = "sfx32gui.dat";
				else if (*p == '=')
					sfxdat = p + 1;
				else
					return FALSE;
			}
			else
			{
				while (*opt)
				{
					switch (*opt)
					{
					case 'b':
						options.fTemp = TRUE;
						if (++i >= ac)
							return FALSE;
						options.szTempDir = av[i];
						break;
					case 'e':
						options.fEncrypt = TRUE;
						break;
					case 'P':
						options.fEncrypt = TRUE;
						if (++i >= ac)
							return FALSE;
						password = av[i];
						break;
					case 'S':
						options.fSystem = TRUE;
						break;
					case '$':
						options.fVolume = TRUE;
						break;
					case 'X':
						options.fExtra = TRUE;
						break;
					case 'D':
						options.fNoDirEntries = TRUE;
						break;
					case 't':
						if (*(opt + 1) == 't')
						{
							options.fExcludeDate = TRUE;
							opt++;
						}
						else
							options.fIncludeDate = TRUE;
						if (++i >= ac)
							return FALSE;
						options.Date = av[i];
						break;
					case 'v':
						options.fVerbose = TRUE;
						break;
					case 'q':
						options.fQuiet = TRUE;
						break;
					case 'l':
						if (*(opt + 1) == 'l')
						{
							options.fCRLF_LF = TRUE;
							opt++;
						}
						else
							options.fLF_CRLF = TRUE;
						break;
					case 'j':
						options.fJunkDir = TRUE;
						break;
					case 'g':
						options.fGrow = TRUE;
						break;
					case 'k':
						options.fForce = TRUE;
						break;
					case 'm':
						options.fMove = TRUE;
						break;
					case 'd':
						options.fDeleteEntries = TRUE;
						break;
					case 'u':
						options.fUpdate = TRUE;
						break;
					case 'f':
						options.fFreshen = TRUE;
						break;
					case 'J':
						options.fJunkSFX = TRUE;
						break;
					case 'o':
						options.fLatestTime = TRUE;
						break;
					case 'z':
						options.fComment = TRUE;
						break;
					case 'A':
						options.fOffsets = TRUE;
						break;
					case '!':
						options.fPrivilege = TRUE;
						break;
					case 'r':
						options.fRecurse = 1;
						break;
					case 'R':
						options.fRecurse = 2;
						break;
					case 'F':
						if (*(opt + 1) == 'F')
						{
							options.fRepair = 2;
							opt++;
						}
						else
							options.fRepair = 1;
						break;
					default:
						if (isdigit (*opt))
							options.fLevel = *opt;
						else
							return FALSE;
						break;
					}
					opt++;
				}
			}
		}
		else
		{
			if (*(av[i] + strlen (av[i]) - 1) == '\\')
			{
				*(av[i] + strlen (av[i]) - 1) = '\0';
				options.szRootDir = av[i];
			}
			else if (archive.lpszZipFN == NULL)
			{
				archive.lpszZipFN = av[i];
			}
			else
			{
				if (archive.argc == FNV_max)
				{
					FNV_max += 32;
					char **x = (char **)realloc (archive.FNV, sizeof *x * FNV_max);
					if (!x)
						return FALSE;
					archive.FNV = x;
				}
				archive.FNV[archive.argc++] = av[i];
			}
		}
	}
	return TRUE;
}

BOOL ZpMakeSFX()
{
	static char dat[PATH_MAX], arc[PATH_MAX], sfx[PATH_MAX];
	FILE *datfp, *arcfp, *sfxfp;
	char buf[1024];
	int in, out;
	
	// dat
	if (!SearchPath (NULL, sfxdat, NULL, sizeof (dat), dat, NULL))
	{
		fprintf (stderr, "%s: t@C܂", sfxdat);
		fcloseall ();
		return FALSE;
	}
	datfp = fopen (dat, "rb");
	if (!datfp)
	{
		fprintf (stderr, "%s: t@CJ܂", sfxdat);
		fcloseall ();
		return FALSE;
	}
	// arc
	if (options.szRootDir)
		SetCurrentDirectory (options.szRootDir);
	strcpy (arc, archive.lpszZipFN);
	arcfp = fopen (arc, "rb");
	if (!arcfp)
	{
		strcat (arc, ".zip");
		arcfp = fopen (arc, "rb");
		if (!arcfp)
		{
			fprintf (stderr, "%s: t@CJ܂", archive.lpszZipFN);
			fcloseall ();
			return FALSE;
		}
	}
	// sfx
	strcpy (sfx, archive.lpszZipFN);
	if (!(strlen (sfx) >= 4 && stricmp (sfx + strlen (sfx) -4 , ".zip") == 0))
		strcat (sfx, ".exe");
	else
		strcpy (sfx + strlen (sfx) - 4, ".exe");
	// ㏑h~
	if (GetFileAttributes (sfx) != 0xFFFFFFFF)
	{
		fprintf (stderr, "%s: ̃t@C܂", sfx);
		fcloseall ();
		return FALSE;
	}
	sfxfp = fopen (sfx, "wb");
	if (!sfxfp)
	{
		fprintf (stderr, "%s: t@CJ܂", sfx);
		fcloseall ();
		return FALSE;
	}
	// dat->sfx
	while ((in = fread (buf, 1, 1024, datfp)) > 0)
	{
		out = fwrite (buf, 1, in, sfxfp);
		if (in != out)
		{
			fprintf (stderr, "dat->sfx s");
			fcloseall ();
			return FALSE;
		}
	}
	// arc->sfx
	while ((in = fread (buf, 1, 1024, arcfp)) > 0)
	{
		out = fwrite (buf, 1, in, sfxfp);
		if (in != out)
		{
			fprintf (stderr, "arc->sfx s");
			fcloseall ();
			return FALSE;
		}
	}
	fcloseall ();
	remove (arc);
	archive.lpszZipFN = sfx;
	options.fOffsets = TRUE;
	return TRUE;
}

int ZpMain(int argc, char **argv)
{
	int ret;
	
	if (!ZpLoadLibrary ())
	{
		fprintf (stderr, "ZIP32.DLL [hs");
		return -1;
	}
	if (!ZpCheckLibrary ())
	{
		fprintf (stderr, "ZIP32.DLL Ver2.3 ȊO̓T|[gĂ܂");
		return -1;
	}
	if (!ZpInitLibrary ())
	{
		fprintf (stderr, "ZIP32.DLL s");
		return -1;
	}
	if (!ZpParseCmdLine (argc, argv))
	{
		fprintf (stderr, "ZIP R}hC͎s");
		return -1;
	}
	if (!archive.lpszZipFN)
	{
		fprintf (stderr, "Usage: [-<options>...] <archive_file_name>[.zip]\n\t\t\t [<directory_name>\\] [<filespec>...]");
	}
	else
	{
		if (sfxdat && !ZpMakeSFX ())
			return FALSE;
		(*fpZpSetOptions) (&options);
		ret = (*fpZpArchive) (archive);
	}
	ZpDiscardCmdLine ();
	ZpFreeLibrary ();
	return ret;
}

int ZpMain(LPTSTR line)
{
	cmdline cl;
	
	if (cl.parse (line, 1))
	{
		fprintf (stderr, "R}hC͎s");
		return -1;
	}
	
	return ZpMain (cl.argc () - 1, cl.argv () + 1);
}
